//
//  CertificateStore.swift
//  Runner
//
//  Created by Lacuna Software on 5/14/20.
//

import Foundation

public class CertificateStore {
    
    public static func listCertificatesWithKey() throws -> [Certificate] {
        var certificates = [Certificate]()
        
        let query: [String: Any] = [
            kSecClass as String: kSecClassIdentity,
            kSecReturnRef as String: true,
            kSecMatchLimit as String: kSecMatchLimitAll
        ]
        var items: CFTypeRef?
        let status = SecItemCopyMatching(query as CFDictionary, &items)
        if status != errSecSuccess {
            if status == errSecItemNotFound {
                return certificates
            }
            throw CrossPkiError.error(ErrorCodes.internalError, "Error listing certificates: \(status)")
        }
        let array = items as! NSArray
        
        for item in array {
            let identity = item as! SecIdentity
            var certificate: SecCertificate?
        
            if SecIdentityCopyCertificate(identity, &certificate) == errSecSuccess {
                certificates.append(Certificate(SecCertificateCopyData(certificate!) as Data))
            }
        }
        
        return certificates
    }
    
    public static func importPkcs12(pkcs12Content: Data, password: String) throws {
        let options = [kSecImportExportPassphrase as String: password]
        var rawItems: CFArray?
        let status = SecPKCS12Import(pkcs12Content as CFData, options as CFDictionary, &rawItems)
        if status != errSecSuccess {
            if status == errSecAuthFailed {
                throw CrossPkiError.error(ErrorCodes.wrongPassword, "The password is incorrect or PKCS#12 data was damaged")
            }
            throw CrossPkiError.error(ErrorCodes.internalError, "Error while importing PKCS#12: \(status)")
        }
        
        let items = rawItems! as! Array<Dictionary<String, Any>>
        for item in items {
            let identity = item[kSecImportItemIdentity as String] as! SecIdentity?
            if identity != nil {
                let addquery: [String: Any] = [
                    kSecValueRef as String: identity!
                ]
                let status = SecItemAdd(addquery as CFDictionary, nil)
                if status != errSecSuccess && status != errSecDuplicateItem {
                    throw CrossPkiError.error(.internalError, "Error importing identity from PKCS#12: \(status)")
                }
            }
        }
    }
    
    public static func importCertificate(certificateContent: Data, keyId: String) throws {
        let certificate = SecCertificateCreateWithData(nil, certificateContent as NSData)
        if certificate == nil {
            throw CrossPkiError.error(.internalError, "Error decoding certificate content")
        }
        
        let addquery: [String: Any] = [
            kSecClass    as String: kSecClassCertificate,
            kSecValueRef as String: certificate!
        ]
        
        let status = SecItemAdd(addquery as CFDictionary, nil)
        if status != errSecSuccess && status != errSecDuplicateItem {
            throw CrossPkiError.error(.internalError, "Error importing certificate to Key Chain: \(status)")
        }
    }
    
    public static func removeCertificate(certificateThumbprint: String) throws {
        let identity = try findCertificate(certificateThumbprint: certificateThumbprint)
        let delQuery: [String: Any] = [
            kSecValueRef as String: identity
        ]
        let status = SecItemDelete(delQuery as CFDictionary)
        if status != errSecSuccess && status != errSecItemNotFound {
            throw CrossPkiError.error(.internalError, "Error removing certificate (\(certificateThumbprint)): \(status)")
        }
    }
    
    public static func findCertificate(certificateThumbprint: String) throws -> SecIdentity {
        let query: [String: Any] = [
            kSecClass      as String: kSecClassIdentity,
            kSecReturnRef  as String: true,
            kSecMatchLimit as String: kSecMatchLimitAll
        ]
        var items: CFTypeRef?
        var status = SecItemCopyMatching(query as CFDictionary, &items)
        if status != errSecSuccess {
            throw CrossPkiError.error(.internalError, "Could not find certificate (\(certificateThumbprint)). Error listing certificates: \(status)")
        }
        let array = items as! NSArray
        
        for item in array {
            let identity = item as! SecIdentity
            var certificate: SecCertificate?
        
            status = SecIdentityCopyCertificate(identity, &certificate)
            if status != errSecSuccess {
                throw CrossPkiError.error(.internalError, "Could not find certificate (\(certificateThumbprint)). Error decoding certificates: \(status)")
            }
            
            if certificateThumbprint == DigestAlgorithm.sha256.computeHash(data: SecCertificateCopyData(certificate!) as Data).base64EncodedString() {
                return identity
            }
        }
        throw CrossPkiError.error(.certificateNotFound, "Certificate not found: \(certificateThumbprint)")
    }
}
