//
//  KeyStore.swift
//  Runner
//
//  Created by Lacuna Software on 5/13/20.
//

import Foundation

public class KeyStore {

    public static func generateRsaKeyPair(keySize: Int) throws -> MobileKeyPair {
        guard #available(iOS 10.0, *) else {
            throw CrossPkiError.error(.algorithmNotSupported, "Public key exporting not supported. Please update to iOS 10+")
        }
        
        let keyId = UUID().uuidString.lowercased().replacingOccurrences(of: "-", with: "")

        let privateKeyDict: [String: Any] = [
            kSecAttrIsPermanent    as String: true,
            kSecAttrApplicationTag as String: formatKeyTag(keyId: keyId).data(using: .utf8)!
        ]

        let keysDict: [String: Any] = [
            kSecAttrType          as String: kSecAttrKeyTypeRSA,
            kSecAttrKeySizeInBits as String: keySize,
            kSecPrivateKeyAttrs   as String: privateKeyDict
        ]
        
        var privateKey: SecKey?
        var publicKey: SecKey?
        var error: Unmanaged<CFError>?
        
        let status = SecKeyGeneratePair(keysDict as CFDictionary, &publicKey, &privateKey)
        if status != errSecSuccess {
            throw CrossPkiError.error(.internalError, "Error generating RSA key pair: \(status)")
        }
        
        let publicKeyData = SecKeyCopyExternalRepresentation(publicKey!, &error)
        if publicKeyData == nil {
            throw CrossPkiError.error(.internalError, "Error exporting public key: \(error!.takeRetainedValue().localizedDescription)")
        }
        
        return MobileKeyPair(keyId: keyId, publicKeyEncoded: publicKeyData! as Data)
    }
    
    public static func getPrivateKey(identity: SecIdentity) throws -> SecKey {
        var privateKey: SecKey?
        let status = SecIdentityCopyPrivateKey(identity, &privateKey)
        if status != errSecSuccess {
            throw CrossPkiError.error(.internalError, "Error while retrieving private key: \(status)")
        }
        return privateKey!
    }
    
    public static func getPrivateKey(keyId: String) throws -> SecKey {
        let query: [String: Any] = [
            kSecClass              as String: kSecClassKey,
            kSecAttrApplicationTag as String: formatKeyTag(keyId: keyId).data(using: .utf8)!,
            kSecMatchLimit         as String: kSecMatchLimitOne,
            kSecReturnRef          as String: true
        ]
        
        var privateKeyRef: CFTypeRef?
        let status = SecItemCopyMatching(query as CFDictionary, &privateKeyRef)
        
        guard status == errSecSuccess else {
            throw CrossPkiError.error(.keyNotFound, "Could not find key (\(keyId): \(status)")
        }
        
        let privateSecKey: SecKey? = (privateKeyRef as! SecKey)
        if privateSecKey == nil {
            throw CrossPkiError.error(.internalError, "Error while getting private key: \(keyId)")
        }
        
        return privateSecKey!
    }
    
    public static func signData(privateKey: SecKey, digestAlgorithm: DigestAlgorithm, data: Data) throws -> Data {
        let hash = digestAlgorithm.computeHash(data: data)
        return try signHash(privateKey: privateKey, digestAlgorithm: digestAlgorithm, hash: hash)
    }
    
    public static func signHash(privateKey: SecKey, digestAlgorithm: DigestAlgorithm, hash: Data) throws -> Data {
        let digestInfo = try DigestInfo.encode(digestAlgorithm: digestAlgorithm, digestValue: hash)
        var signatureBufferLength = 2 * 1024
        var signature = [UInt8](repeating: 0,  count: signatureBufferLength)
        
        var digestInfoBuffer = [UInt8](digestInfo)
        
        let status = SecKeyRawSign(privateKey, SecPadding.PKCS1, &digestInfoBuffer, digestInfo.count, &signature, &signatureBufferLength)
        
        if status != errSecSuccess {
            throw CrossPkiError.error(.internalError, "Error while signing hash: \(status)")
        }
        
        return Data(bytes: signature, count: signatureBufferLength)
    }

    private static func formatKeyTag(keyId: String) -> String {
        let bundleId = Bundle.main.bundleIdentifier
        return "\(bundleId!).keys.\(keyId)"
    }
}
