import Flutter
import UIKit

public class SwiftCrossPkiPlugin: NSObject, FlutterPlugin {

	public static func register(with registrar: FlutterPluginRegistrar) {
		let channel = FlutterMethodChannel(name: "plugin.lacunasoftware.com/crosspki", binaryMessenger: registrar.messenger())
		let instance = SwiftCrossPkiPlugin()
		registrar.addMethodCallDelegate(instance, channel: channel)
	}

	public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        do {
            let arguments = call.arguments as? NSDictionary

            switch call.method {
                case "listCertificatesWithKey":
                    result(try listCertificatesWithKey())
                    
                case "importPkcs12":
                    try importPkcs12(arguments)
                    result(nil)
                
                case "importCertificate":
                    try importCertificate(arguments)
                    result(nil)
                    
                case "removeCertificate":
                    try removeCertificate(arguments)
                    result(nil)
                    
                case "generateRsaKeyPair":
                    result(try generateRsaKeyPair(arguments))
                
                case "signData":
                    result(try signData(arguments))
                
                case "signHash":
                    result(try signHash(arguments))
                
                case "keySignData":
                    result(try keySignData(arguments))
                
                case "keySignHash":
                    result(try keySignHash(arguments))
                    
                default:
                    result(nil)
            }
            
        } catch CrossPkiError.error(let code, let message) {
            result(FlutterError(code: "\(code)", message: message, details: nil))
            
        } catch {
            result(FlutterError(code: "\(ErrorCodes.unknown)", message: "\(error)", details: nil))
        }
	}
    
    private func listCertificatesWithKey() throws -> [[String: Any]] {
        let certificates = try CertificateStore.listCertificatesWithKey()
        return Converters.toDictionary(certificates)
    }
    
    private func removeCertificate(_ arguments: NSDictionary?) throws {
        let certificateThumbprint = arguments?["certificateThumbprint"] as? String
        try CertificateStore.removeCertificate(certificateThumbprint: certificateThumbprint!)
    }
    
    private func importPkcs12(_ arguments: NSDictionary?) throws {
        let pkcs12Content = arguments?["pkcs12Content"] as? FlutterStandardTypedData
        let password = arguments?["password"] as? String
        try CertificateStore.importPkcs12(pkcs12Content: pkcs12Content!.data, password: password!)
    }
    
    private func importCertificate(_ arguments: NSDictionary?) throws {
        let certificateContent = arguments?["certificateContent"] as? FlutterStandardTypedData
        let keyId = arguments?["keyId"] as? String
        try CertificateStore.importCertificate(certificateContent: certificateContent!.data, keyId: keyId!)
    }
    
    private func generateRsaKeyPair(_ arguments: NSDictionary?) throws -> [String: Any] {
        let keySize = arguments?["keySize"] as? Int
        let keyPair = try KeyStore.generateRsaKeyPair(keySize: keySize!)
        return Converters.toDictionary(keyPair)
    }
    
    private func signData(_ arguments: NSDictionary?) throws -> [String: Any] {
        let certificateThumbprint = arguments?["certificateThumbprint"] as? String
        let identity = try CertificateStore.findCertificate(certificateThumbprint: certificateThumbprint!)
        let privateKey = try KeyStore.getPrivateKey(identity: identity)
        return try commonSignData(arguments, privateKey)
    }

    private func signHash(_ arguments: NSDictionary?) throws -> [String: Any] {
        let certificateThumbprint = arguments?["certificateThumbprint"] as? String
        let identity = try CertificateStore.findCertificate(certificateThumbprint: certificateThumbprint!)
        let privateKey = try KeyStore.getPrivateKey(identity: identity)
        return try commonSignHash(arguments, privateKey)
    }

    private func keySignData(_ arguments: NSDictionary?) throws -> [String: Any] {
        let keyId = arguments?["keyId"] as? String
        let privateKey = try KeyStore.getPrivateKey(keyId: keyId!)
        return try commonSignData(arguments, privateKey)
    }

    private func keySignHash(_ arguments: NSDictionary?) throws -> [String: Any] {
        let keyId = arguments?["keyId"] as? String
        let privateKey = try KeyStore.getPrivateKey(keyId: keyId!)
        return try commonSignHash(arguments, privateKey)
    }
    
    private func commonSignData(_ arguments: NSDictionary?, _ key: SecKey) throws -> [String: Any] {
        let digestAlgorithmStr = arguments?["digestAlgorithm"] as? String
        let data = arguments?["data"] as? FlutterStandardTypedData
        let digestAlgorithm = try DigestAlgorithm.parse(digestAlgorithmStr!)
        let signature = try KeyStore.signData(privateKey: key, digestAlgorithm: digestAlgorithm, data: data!.data)
        return Converters.toSignatureDictionary(signature)
    }
    
    private func commonSignHash(_ arguments: NSDictionary?, _ key: SecKey) throws -> [String: Any] {
        let digestAlgorithmStr = arguments?["digestAlgorithm"] as? String
        let hash = arguments?["hash"] as? FlutterStandardTypedData
        let digestAlgorithm = try DigestAlgorithm.parse(digestAlgorithmStr!)
        let signature = try KeyStore.signHash(privateKey: key, digestAlgorithm: digestAlgorithm, hash: hash!.data)
        return Converters.toSignatureDictionary(signature)
    }
}
