import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'dart:typed_data';

import 'package:asn1lib/asn1lib.dart';
import 'package:flutter/services.dart';

/// Cross PKI Flutter Plugin
class CrossPki {
  static const MethodChannel _channel =
      const MethodChannel('plugin.lacunasoftware.com/crosspki');

  /// Lists the certificates with associated private key from the platform native secure Key Store.
  /// Throws a [CrossPkiException] on any platform error.
  static Future<List<PKCertificate>> listCertificatesWithKey() async {
    List<dynamic> certificates = await _handledInvoke('listCertificatesWithKey');
    return certificates
      .cast<Map<dynamic, dynamic>>()
      .map((c) => PKCertificate._fromMap(c))
      .toList();
  }

  /// Imports a PFX (PKCS#12) bundle with certificates and private keys to the platform native secure Key Store.
  /// Throws a [CrossPkiException] on any platform error.
  static Future<void> importPkcs12(Uint8List pkcs12Content, String password) async {
    await _handledInvoke('importPkcs12', <String, dynamic> {
      'pkcs12Content': pkcs12Content,
      'password': password,
    });
  }

  /// Imports a certificate issued to a previous generated key pair using [generateRsaKeyPair].
  /// Throws a [CrossPkiException] on any platform error.
  static Future<void> importCertificate(Uint8List certificateContent, String keyId) async {
    await _handledInvoke('importCertificate', <String, dynamic> {
      'certificateContent': certificateContent,
      'keyId': keyId,
    });
  }

  /// Removes a certificate with key entry from platform native secure Key Store.
  ///
  /// [certificateThumbprint] parameter is retrieved on [PKCertificate.thumbprint] using [listCertificatesWithKey] method.
  /// Throws a [CrossPkiException] on any platform error.
  static Future<void> removeCertificate(String certificateThumbprint) async {
    await _handledInvoke('removeCertificate', <String, dynamic> {
      'certificateThumbprint': certificateThumbprint,
    });
  }

  /// Generates a RSA key pair on the platform native secure Key Store.
  /// Throws a [CrossPkiException] on any platform error.
  static Future<MobileKeyPair> generateRsaKeyPair(int keySize) async {
    Map<dynamic, dynamic> keyPair = await _handledInvoke('generateRsaKeyPair', <String, dynamic> {
      'keySize': keySize,
    });
    return MobileKeyPair._fromMap(keyPair, PKAlgorithm.rsa);
  }

  /// Signs data using a selected certificate's private key.
  ///
  /// [certificateThumbprint] parameter is retrieved on [PKCertificate.thumbprint] using [listCertificatesWithKey] method.
  /// Throws a [CrossPkiException] on any platform error.
  static Future<Uint8List> signData(String certificateThumbprint, DigestAlgorithm digestAlgorithm, Uint8List data) async {
    Map<dynamic, dynamic> signatureResponse = await _handledInvoke('signData', <String, dynamic> {
      'certificateThumbprint': certificateThumbprint,
      'digestAlgorithm': digestAlgorithm.oid,
      'data': data,
    });
    return signatureResponse["signature"];
  }

  /// Signs a hash using a selected certificate's private key.
  ///
  /// [certificateThumbprint] parameter is retrieved on [PKCertificate.thumbprint] using [listCertificatesWithKey] method.
  /// Throws a [CrossPkiException] on any platform error.
  static Future<Uint8List> signHash(String certificateThumbprint, DigestAlgorithm digestAlgorithm, Uint8List hash) async {
    Map<dynamic, dynamic> signatureResponse = await _handledInvoke('signHash', <String, dynamic> {
      'certificateThumbprint': certificateThumbprint,
      'digestAlgorithm': digestAlgorithm.oid,
      'hash': hash,
    });
    return signatureResponse["signature"];
  }

  /// Signs data using the private key of a previous generated key pair using method [generateRsaKeyPair].
  ///
  /// [certificateThumbprint] parameter is retrieved on [PKCertificate.thumbprint] using [listCertificatesWithKey] method.
  /// Throws a [CrossPkiException] on any platform error.
  static Future<Uint8List> keySignData(String keyId, DigestAlgorithm digestAlgorithm, Uint8List data) async {
    Map<dynamic, dynamic> signatureResponse = await _handledInvoke('keySignData', <String, dynamic> {
      'keyId': keyId,
      'digestAlgorithm': digestAlgorithm.oid,
      'data': data,
    });
    return signatureResponse["signature"];
  }

  /// Signs a hash using the private key of a previous generated key pair using method [generateRsaKeyPair].
  ///
  /// [certificateThumbprint] parameter is retrieved on [PKCertificate.thumbprint] using [listCertificatesWithKey] method.
  /// Throws a [CrossPkiException] on any platform error.
  static Future<Uint8List> keySignHash(String keyId, DigestAlgorithm digestAlgorithm, Uint8List hash) async {
    Map<dynamic, dynamic> signatureResponse = await _handledInvoke('keySignHash', <String, dynamic> {
      'keyId': keyId,
      'digestAlgorithm': digestAlgorithm.oid,
      'hash': hash,
    });
    return signatureResponse["signature"];
  }

  /// Generates a Certificate Signing Request (CSR / PKCS#10) DER encoded for a generated key pair using [generateRsaKeyPair].
  /// Throws a [CrossPkiException] on any platform error.
  static Future<Uint8List> generateCsr(MobileKeyPair keyPair) async {
    var signatureAlgorithm = keyPair.algorithm.getSignatureAlgorithm(DigestAlgorithm.sha256);
    var crInfo = _Asn1Util.createCertificationRequestInfo(keyPair.publicKeyEncoded, keyPair.algorithm);
    var signature = await CrossPki.keySignData(keyPair.keyId, signatureAlgorithm.digestAlgorithm, crInfo.encodedBytes);
    return _Asn1Util.createCsr(crInfo, signatureAlgorithm, signature);
  }

  static Future<T> _handledInvoke<T>(String method, [dynamic arguments]) async {
    try {
      return await _channel.invokeMethod(method, arguments);
    } on PlatformException catch (e) {
      throw CrossPkiException._(e.code, e.message);
    }
  }
}

//-----------------------------------------------------------------------------
// Exception

enum CrossPkiErrorCodes {
  unknown,
  internal,
  invalidLicense,
  wrongPassword,
  platformNotSupported,
  algorithmNotSupported,
  unknownAlgorithm,
  certificateNotFound,
  keyNotFound,
}

class CrossPkiException implements Exception {

  CrossPkiErrorCodes _code;
  CrossPkiErrorCodes get code => _code;

  final String message;

  CrossPkiException._(String codeStr, this.message) {
    _code = CrossPkiErrorCodes.values.firstWhere((c) => c.toString() == 'CrossPkiErrorCodes.$codeStr', orElse: () => CrossPkiErrorCodes.unknown);
  }

  CrossPkiException(CrossPkiErrorCodes code, this.message) {
    _code = code;
  }

  @override
  String toString() => '[$code]: $message';
}

//-----------------------------------------------------------------------------
// Models

class PKCertificate {

  /// A unique identifier for the user certificate.
  final String thumbprint;

  /// Certificate Subject DN name entries.
  final Name subjectName;

  /// Certificate Issuer DN name entries.
  final Name issuerName;

  /// A summary display name for certificate Subject.
  final String subjectDisplayName;

  /// A summary display name for certificate Issuer.
  final String issuerDisplayName;

  /// Certificate validity starting DateTime.
  final DateTime validityStart;

  /// Certificate validity ending DateTime.
  final DateTime validityEnd;

  /// Certificate DER encoded bytes.
  final Uint8List encoded;

  PKCertificate._({
    this.thumbprint,
    this.subjectName,
    this.issuerName,
    this.subjectDisplayName,
    this.issuerDisplayName,
    this.validityStart,
    this.validityEnd,
    this.encoded,
  });

  static PKCertificate _fromMap(Map map) {
    return _Asn1Util.decodeCertificate(map['encoded'], map['thumbprint']);
  }
}

/// X509 Distinguished Name class.
class Name {

  /// X.520 id-at-commonName
  final String commonName;

  /// X.520 id-at-surname
  final String surname;

  /// X.520 id-at-serialNumber
  final String serialNumber;

  /// X.520 id-at-countryName
  final String country;

  /// X.520 id-at-localityName
  final String locality;

  /// X.520 id-at-stateName
  final String stateName;

  /// X.520 id-at-streetAddress
  final String streetAddress;

  /// X.520 id-at-organizationName
  final String organization;

  /// X.520 id-at-organizationalUnitName
  final String organizationalUnit;

  /// X.520 id-at-title
  final String title;

  /// X.520 id-at-businessCategory
  final String businessCategory;

  /// X.520 id-at-postalCode
  final String postalCode;

  /// X.520 id-at-telephoneNumber
  final String telephoneNumber;

  /// X.520 id-at-givenName
  final String givenName;

  /// X.520 id-at-initials
  final String initials;

  /// X.520 id-at-generationQualifier
  final String generationQualifier;

  /// X.520 id-at-dnQualifier
  final String dnQualifier;

  /// X.520 at-pseudonym
  final String pseudonym;

  /// PKCS #9 pkcs-9-at-emailAddress
  final String emailAddress;

  /// LDAP RFC 4519 uid / RFC 1274 userid
  final String userId;

  /// The name struct ala RFC 2253
  final String stringValue;

  Name._({
    this.commonName,
    this.surname,
    this.serialNumber,
    this.country,
    this.locality,
    this.stateName,
    this.streetAddress,
    this.organization,
    this.organizationalUnit,
    this.title,
    this.businessCategory,
    this.postalCode,
    this.telephoneNumber,
    this.givenName,
    this.initials,
    this.generationQualifier,
    this.dnQualifier,
    this.pseudonym,
    this.emailAddress,
    this.userId,
    this.stringValue,
  });

  @override
  String toString() => stringValue;

  static Name _fromMap(Map map, String stringValue) {
    return Name._(
      commonName: map[_X509NamesOids.cn],
      surname: map[_X509NamesOids.surname],
      serialNumber: map[_X509NamesOids.serialNumber],
      country: map[_X509NamesOids.c],
      locality: map[_X509NamesOids.l],
      stateName: map[_X509NamesOids.st],
      streetAddress: map[_X509NamesOids.street],
      organization: map[_X509NamesOids.o],
      organizationalUnit: map[_X509NamesOids.ou],
      title: map[_X509NamesOids.t],
      businessCategory: map[_X509NamesOids.businessCategory],
      postalCode: map[_X509NamesOids.postalCode],
      telephoneNumber: map[_X509NamesOids.telephoneNumber],
      givenName: map[_X509NamesOids.givenName],
      initials: map[_X509NamesOids.initials],
      generationQualifier: map[_X509NamesOids.generation],
      dnQualifier: map[_X509NamesOids.dnQualifier],
      pseudonym: map[_X509NamesOids.pseudonym],
      emailAddress: map[_X509NamesOids.e],
      userId: map[_X509NamesOids.uid],
      stringValue: stringValue,
    );
  }
}

// ----------------------------------------------------------------------------

class MobileKeyPair {

  /// Public Key (Asymmetric) algorithm type.
  final PKAlgorithm algorithm;

  /// Generated private key Id.
  final String keyId;

  /// Generated public key DER encoded bytes.
  final Uint8List publicKeyEncoded;

  MobileKeyPair._({
    this.keyId,
    this.publicKeyEncoded,
    this.algorithm,
  });

  static MobileKeyPair _fromMap(Map map, PKAlgorithm algorithm) {
    return MobileKeyPair._(
      keyId: map['keyId'],
      publicKeyEncoded: _Asn1Util.getPublicKeyEncoded(map['publicKeyEncoded']),
      algorithm: algorithm,
    );
  }
}

// ----------------------------------------------------------------------------
// Algorithms

/// Base algorithms class.
abstract class BaseAlgorithm {

  /// Algorithm name.
  final String name;

  /// Algorithm OID.
  final String oid;

  const BaseAlgorithm(this.name, this.oid);

  @override
  String toString() => '$name ($oid)';

  @override
  bool operator ==(other) => other is BaseAlgorithm && oid == other.oid;

  static String _canonString(String value) {
    return value.toLowerCase().trim().replaceAll('-', '').replaceAll(' ', '');
  }

  @override
  int get hashCode => super.hashCode;

}

// ----------------------------------------------------------------------------

/// Public Key (Asymmetric) algorithms class.
class PKAlgorithm extends BaseAlgorithm {

  const PKAlgorithm._(String name, String oid): super(name, oid);

  static const rsa = const PKAlgorithm._('RSA', _PkiOids.rsa);

  SignatureAlgorithm getSignatureAlgorithm(DigestAlgorithm digestAlgorithm) {
    return SignatureAlgorithm._allAlgorithms.firstWhere((a) => a.pkAlgorithm == this && a.digestAlgorithm == digestAlgorithm);
  }
}

// ----------------------------------------------------------------------------

/// Digest (Hash) algorithms class.
class DigestAlgorithm extends BaseAlgorithm {

  const DigestAlgorithm._(String name, String oid): super(name, oid);

  static const md5 = const DigestAlgorithm._('MD5', _PkiOids.md5);
  static const sha1 = const DigestAlgorithm._('SHA-1', _PkiOids.sha1);
  static const sha224 = const DigestAlgorithm._('SHA-224', _PkiOids.sha224);
  static const sha256 = const DigestAlgorithm._('SHA-256', _PkiOids.sha256);
  static const sha384 = const DigestAlgorithm._('SHA-384', _PkiOids.sha384);
  static const sha512 = const DigestAlgorithm._('SHA-512', _PkiOids.sha512);

  static List<DigestAlgorithm> _allAlgorithms = [
    md5, sha1, sha224, sha256, sha384, sha512
  ];

  /// Parses a digest (hash) algorithm name or OID to [DigestAlgorithm] object.
  /// Throws [CrossPkiException] on any unknown or not supported algorithm.
  static DigestAlgorithm fromString(String value) {
    final d = _allAlgorithms.firstWhere((element) => element.oid == value || BaseAlgorithm._canonString(value) == BaseAlgorithm._canonString(element.name), orElse: () => null);
    if (d == null) {
      throw CrossPkiException(CrossPkiErrorCodes.unknownAlgorithm, 'Could not parse digest algorithm: $value');
    }
    return d;
  }
}

// ----------------------------------------------------------------------------

/// Signature algorithms class.
class SignatureAlgorithm extends BaseAlgorithm {

  final PKAlgorithm pkAlgorithm;
  final DigestAlgorithm digestAlgorithm;

  const SignatureAlgorithm._(this.pkAlgorithm, this.digestAlgorithm, String name, String oid): super(name, oid);

  static const md5WithRsa = SignatureAlgorithm._(PKAlgorithm.rsa, DigestAlgorithm.md5, 'MD5withRSA', _PkiOids.md5WithRsa);
  static const sha1WithRsa = SignatureAlgorithm._(PKAlgorithm.rsa, DigestAlgorithm.sha1, 'SHA1withRSA', _PkiOids.sha1WithRsa);
  static const sha224WithRsa = SignatureAlgorithm._(PKAlgorithm.rsa, DigestAlgorithm.sha224, 'SHA224withRSA', _PkiOids.sha224WithRsa);
  static const sha256WithRsa = SignatureAlgorithm._(PKAlgorithm.rsa, DigestAlgorithm.sha256, 'SHA256withRSA', _PkiOids.sha256WithRsa);
  static const sha384WithRsa = SignatureAlgorithm._(PKAlgorithm.rsa, DigestAlgorithm.sha384, 'SHA384withRSA', _PkiOids.sha384WithRsa);
  static const sha512WithRsa = SignatureAlgorithm._(PKAlgorithm.rsa, DigestAlgorithm.sha512, 'SHA512withRSA', _PkiOids.sha512WithRsa);

  static List<SignatureAlgorithm> _allAlgorithms = [
    md5WithRsa, sha1WithRsa, sha224WithRsa, sha256WithRsa, sha384WithRsa, sha512WithRsa
  ];
}

// ----------------------------------------------------------------------------

class _Asn1Util {

  static ASN1Sequence createCertificationRequestInfo(Uint8List publicKey, PKAlgorithm publicKeyAlgorithm) {
    var certificationRequestInfo = ASN1Sequence();
    // version
    certificationRequestInfo.add(ASN1Integer.fromInt(0));
    // subject
    certificationRequestInfo.add(ASN1Object.fromBytes(_subjectNameEncoded));
    // subjectPKInfo
    certificationRequestInfo.add(createSubjectPublicKeyInfo(publicKey, publicKeyAlgorithm));
    // attributes
    certificationRequestInfo.add(ASN1Null(tag: 0xA0));
    return certificationRequestInfo;
  }

  static Uint8List createCsr(ASN1Sequence certificationRequestInfo, SignatureAlgorithm signatureAlgorithm, Uint8List signature) {
    var csr = ASN1Sequence();
    // certificationRequestInfo;
    csr.add(certificationRequestInfo);
    // signatureAlgorithm
    csr.add(createAlgorithmIdentifier(signatureAlgorithm));
    // signature
    csr.add(ASN1BitString(signature.toList()));
    return csr.encodedBytes;
  }

  static ASN1Sequence createSubjectPublicKeyInfo(Uint8List publicKey, PKAlgorithm publicKeyAlgorithm) {
    var subjectPublicKeyInfo = ASN1Sequence();
    // algorithm
    subjectPublicKeyInfo.add(createAlgorithmIdentifier(publicKeyAlgorithm));
    // subjectPublicKey
    subjectPublicKeyInfo.add(ASN1BitString(publicKey.toList()));
    return subjectPublicKeyInfo;
  }

  static ASN1Sequence createAlgorithmIdentifier(BaseAlgorithm algorithm) {
    var algorithmSeq = ASN1Sequence();
    algorithmSeq.add(ASN1ObjectIdentifier.fromComponentString(algorithm.oid));

    if (algorithm == PKAlgorithm.rsa || (algorithm is SignatureAlgorithm && algorithm.pkAlgorithm == PKAlgorithm.rsa)) {
      algorithmSeq.add(ASN1Null(tag: 0x05));
      return algorithmSeq;
    }

    throw CrossPkiException(CrossPkiErrorCodes.algorithmNotSupported, 'Algorithm not supported: $algorithm');
  }

  static Uint8List getPublicKeyEncoded(Uint8List platformPublicKeyResult) {
    if (Platform.isIOS) {
      return platformPublicKeyResult;
    }

    if (Platform.isAndroid) {
      var subjectPublicKeyInfo = ASN1Sequence.fromBytes(platformPublicKeyResult);
      var bitString = subjectPublicKeyInfo.elements[1] as ASN1BitString;
      return bitString.contentBytes();
    }

    throw CrossPkiException(CrossPkiErrorCodes.platformNotSupported, 'Platform not supported: ${Platform.operatingSystem}');
  }

  static PKCertificate decodeCertificate(Uint8List content, String thumbprint) {
    var tbsCertificate = ASN1Sequence.fromBytes(content).elements.first as ASN1Sequence;
    var issuerDN = _Asn1Util.decodeName(tbsCertificate.elements[3] as ASN1Sequence);
    var validity = _Asn1Util.decodeValidity(tbsCertificate.elements[4] as ASN1Sequence);
    var subjectDN = _Asn1Util.decodeName(tbsCertificate.elements[5] as ASN1Sequence);

    return PKCertificate._(
      thumbprint: thumbprint,
      subjectName: subjectDN,
      issuerName: issuerDN,
      subjectDisplayName: subjectDN.commonName,
      issuerDisplayName: issuerDN.commonName,
      validityStart: validity['notBefore'],
      validityEnd: validity['notAfter'],
      encoded: content
    );
  }

  static Name decodeName(ASN1Sequence sequence) {
    Map<String, String> name = {};
    List<String> nameString = [];
    sequence.elements.forEach((ASN1Object asn1Set) {
      if (asn1Set is ASN1Set) {
        for (var setSequence in asn1Set.elements) {
          try {
            var key = ((setSequence as ASN1Sequence).elements[0] as ASN1ObjectIdentifier).identifier;
            var value = getStringValue(((setSequence as ASN1Sequence).elements[1]));
            nameString.add('${_X509NamesOids.codes[key] ?? key}=$value');
            if (name.containsKey(key)) {
              value = '${name[key]}, $value';
            }
            name[key] = value;
          } catch (e) { }
        }
      }
    });

    return Name._fromMap(name, nameString.reversed.join(','));
  }

  static Map<String, DateTime> decodeValidity(ASN1Sequence validity) {
    return {
      'notBefore': (validity.elements[0] as ASN1UtcTime).dateTimeValue,
      'notAfter': (validity.elements[1] as ASN1UtcTime).dateTimeValue,
    };
  }

  static String getStringValue(ASN1Object obj) {
    if (obj is ASN1PrintableString) return obj.stringValue;
    if (obj is ASN1IA5String) return obj.stringValue;
    if (obj is ASN1UTF8String) return obj.utf8StringValue;
    throw ASN1Exception('Object $obj is not a ASN1 string type');
  }

  static Uint8List _subjectNameEncoded = base64Decode("MB8xHTAbBgNVBAMTFENyb3NzUGtpLUZsdXR0ZXItS2V5");
}

class _PkiOids {

// Digest Algorithms
  static const md5    = '1.2.840.113549.2.5';
  static const sha1   = '1.3.14.3.2.26';
  static const sha256 = '2.16.840.1.101.3.4.2.1';
  static const sha384 = '2.16.840.1.101.3.4.2.2';
  static const sha512 = '2.16.840.1.101.3.4.2.3';
  static const sha224 = '2.16.840.1.101.3.4.2.4';

// Signature Algorithms
//  static const md2WithRsa    = '1.2.840.113549.1.1.2';
  static const md5WithRsa    = '1.2.840.113549.1.1.4';
  static const sha1WithRsa   = '1.2.840.113549.1.1.5';
  static const sha256WithRsa = '1.2.840.113549.1.1.11';
  static const sha384WithRsa = '1.2.840.113549.1.1.12';
  static const sha512WithRsa = '1.2.840.113549.1.1.13';
  static const sha224WithRsa = '1.2.840.113549.1.1.14';

//  static const sha1WithDsa   = '1.2.840.10040.4.3';
//  static const sha224WithDsa = '2.16.840.1.101.3.4.3.1';
//  static const sha256WithDsa = '2.16.840.1.101.3.4.3.2';

// ANSI X9.62 Elliptic Curve Digital Signature Algorithm (ECDSA) signatures and modules
//  static const ecDsaPrefix     = '1.2.840.10045';
//  static const ecDsaWithSHA1   = '1.2.840.10045.4.1';
//  static const ecDsaWithSHA224 = '1.2.840.10045.4.3.1';
//  static const ecDsaWithSHA256 = '1.2.840.10045.4.3.2';
//  static const ecDsaWithSHA384 = '1.2.840.10045.4.3.3';
//  static const ecDsaWithSHA512 = '1.2.840.10045.4.3.4';

// Asymmetric Algorithms
  static const rsa = '1.2.840.113549.1.1.1';
//  static const dsa = '1.2.840.10040.4.1';
//  static const ec  = '1.2.840.10045.2.1';

}

class _X509NamesOids {
  static const c                    = '2.5.4.6';
  static const o                    = '2.5.4.10';
  static const ou                   = '2.5.4.11';
  static const t                    = '2.5.4.12';
  static const cn                   = '2.5.4.3';
  static const street               = '2.5.4.9';
  static const serialNumber         = '2.5.4.5';
  static const l                    = '2.5.4.7';
  static const st                   = '2.5.4.8';
  static const surname              = '2.5.4.4';
  static const givenName            = '2.5.4.42';
  static const initials             = '2.5.4.43';
  static const generation           = '2.5.4.44';
//  static const uniqueIdentifier     = '2.5.4.45';
  static const businessCategory     = '2.5.4.15';
  static const postalCode           = '2.5.4.17';
  static const dnQualifier          = '2.5.4.46';
  static const pseudonym            = '2.5.4.65';
//  static const dateOfBirth          = '1.3.6.1.5.5.7.9.1';
//  static const placeOfBirth         = '1.3.6.1.5.5.7.9.2';
//  static const gender               = '1.3.6.1.5.5.7.9.3';
//  static const countryOfCitizenship = '1.3.6.1.5.5.7.9.4';
//  static const countryOfResidence   = '1.3.6.1.5.5.7.9.5';
//  static const nameAtBirth          = '1.3.36.8.3.14';
//  static const postalAddress        = '2.5.4.16';
//  static const dmdName              = '2.5.4.54';
  static const telephoneNumber      = '2.5.4.20';
  static const e                    = '1.2.840.113549.1.9.1';
//  static const dc                   = '0.9.2342.19200300.100.1.25';
  static const uid                  = '0.9.2342.19200300.100.1.1';

  static const Map<String, String> codes = {
    '2.5.4.6':                    'C',
    '2.5.4.10':                   'O',
    '2.5.4.11':                   'OU',
    '2.5.4.12':                   'T',
    '2.5.4.3':                    'CN',
    '2.5.4.9':                    'STREET',
    '2.5.4.5':                    'SERIALNUMBER',
    '2.5.4.7':                    'L',
    '2.5.4.8':                    'ST',
    '2.5.4.4':                    'SURNAME',
    '2.5.4.42':                   'GIVENNAME',
    '2.5.4.43':                   'INITIALS',
    '2.5.4.44':                   'GENERATION',
    '2.5.4.45':                   'UniqueIdentifier',
    '2.5.4.15':                   'BusinessCategory',
    '2.5.4.17':                   'PostalCode',
    '2.5.4.46':                   'DN',
    '2.5.4.65':                   'Pseudonym',
    '1.3.6.1.5.5.7.9.1':          'DateOfBirth',
    '1.3.6.1.5.5.7.9.2':          'PlaceOfBirth',
    '1.3.6.1.5.5.7.9.3':          'Gender',
    '1.3.6.1.5.5.7.9.4':          'CountryOfCitizenship',
    '1.3.6.1.5.5.7.9.5':          'CountryOfResidence',
    '1.3.36.8.3.14':              'NameAtBirth',
    '2.5.4.16':                   'PostalAddress',
    '2.5.4.20':                   'TelephoneNumber',
    '1.2.840.113549.1.9.1':       'E',
    '0.9.2342.19200300.100.1.25': 'DC',
    '0.9.2342.19200300.100.1.1':  'UID',
  };
}